import pytest

from smllib.errors import WrongValueType
from smllib.sml_frame_snippet import SmlFrameSnippet


def test_snippet_type():
    s = SmlFrameSnippet(8, 0)

    with pytest.raises(WrongValueType) as e:
        s.get_value(float)
    assert str(e.value) == 'Expected type float but got type int'

    with pytest.raises(WrongValueType) as e:
        s.get_value(list)
    assert str(e.value) == 'Expected type list but got type int'

    s = SmlFrameSnippet([8], 0)

    with pytest.raises(WrongValueType) as e:
        s.get_value(float)
    assert str(e.value) == 'Expected type float but got type list'



from binascii import a2b_hex

from smllib.builder import create_context, SmlListEntryBuilder
from smllib.sml_frame import SmlFrame


def test_sml_fields():
    f = SmlFrame(a2b_hex('77078181c78203ff010101010449534b0177070100000009ff010101010b'))
    val_list = f._parse_msg(f.get_value(0))
    o = SmlListEntryBuilder().build(val_list, create_context())
    assert o.obis == '8181c78203ff'
    assert o.value == 'ISK'

    f = SmlFrame(a2b_hex('77070100010800ff650000018201621e52ff590000000001122334017707'))
    val_list = f._parse_msg(f.get_value(0))
    o = SmlListEntryBuilder().build(val_list, create_context())

    assert o.obis == '0100010800ff'
    assert o.status == 386
    assert o.unit == 30
    assert o.scaler == -1
    assert o.value == 17965876
    assert o.get_value() == 1796587.6


def test_val_time():
    # Frame where time is None
    f = SmlFrame(a2b_hex('77070100600100ff010101010b0a01484c5902000424a001'))
    val_list = f._parse_msg(f.get_value(0))
    o = SmlListEntryBuilder().build(val_list, create_context())
    assert o.val_time is None

    # Frame where secIndex == 1 and time == 0
    # -> 7262016200
    f = SmlFrame(a2b_hex('77070100600100ff017262016200620052000b0a01445a47000282c0b001'))
    val_list = f._parse_msg(f.get_value(0))
    o = SmlListEntryBuilder().build(val_list, create_context())
    assert o.val_time == 0



UNITS = {
    1: 'a',
    2: 'mo',
    3: 'wk',
    4: 'd',
    5: 'h',
    6: 'min',
    7: 's',
    8: '°',
    9: '°C',
    10: 'currency',
    11: 'm',
    12: 'm/s',
    13: 'm³',
    14: 'm³',
    15: 'm³/h',
    16: 'm³/h',
    17: 'm³/d',
    18: 'm³/d',
    19: 'l',
    20: 'kg',
    21: 'N',
    22: 'Nm',
    23: 'Pa',
    24: 'bar',
    25: 'J',
    26: 'J/h',
    27: 'W',
    28: 'VA',
    29: 'var',
    30: 'Wh',
    31: 'VAh',
    32: 'varh',
    33: 'A',
    34: 'C',
    35: 'V',
    36: 'V/m',
    37: 'F',
    38: 'Ω',
    39: 'Ωm²/m',
    40: 'Wb',
    41: 'T',
    42: 'A/m',
    43: 'H',
    44: 'Hz',
    45: '1/(Wh)',
    46: '1/(varh)',
    47: '1/(VAh)',
    48: 'V²h',
    49: 'A²h',
    50: 'kg/s',
    51: 'S',
    52: 'K',
    53: '1/(V²h)',
    54: '1/(A²h)',
    55: '1/m³',
    56: '%',
    57: 'Ah',
    60: 'Wh/m³',
    61: 'J/m³',
    62: 'Mol%',
    63: 'g/m³',
    64: 'Pas',
    65: 'J/kg',
    66: 'g/cm²',
    67: 'arm',
    70: 'dBm',
    71: 'dBµV',
    72: 'dB',
}

# https://www.promotic.eu/en/pmdoc/Subsystems/Comm/PmDrivers/IEC62056_OBIS.htm
OBIS_NAMES = {
    '0100000009ff': 'Geräteeinzelidentifikation',
    '0100010800ff': 'Zählerstand Bezug Total',
    '0100010801ff': 'Zählerstand Bezug Tarif 1',
    '0100010802ff': 'Zählerstand Bezug Tarif 2',
    '0100010803ff': 'Zählerstand Bezug Tarif 3',
    '0100010804ff': 'Zählerstand Bezug Tarif 4',
    '0100011100ff': 'Total-Zählerstand',
    '0100020800ff': 'Zählerstand Einspeisung Total',
    '0100020801ff': 'Zählerstand Einspeisung Tarif 1',
    '0100020802ff': 'Zählerstand Einspeisung Tarif 2',
    '0100020803ff': 'Zählerstand Einspeisung Tarif 3',
    '0100020804ff': 'Zählerstand Einspeisung Tarif 4',
    '01000b0600ff': 'Maximalstrom Total',
    '01000b0700ff': 'Momentanstrom Total',
    '01000c0700ff': 'Momentanspannung Total',
    '01000e0700ff': 'Netzfrequenz',
    '0100100700ff': 'aktuelle Wirkleistung',
    '0100150700ff': 'Momentanwirkleistung Bezug L1',
    '0100160700ff': 'Momentanwirkleistung Einspeisung L1',
    '0100170700ff': 'Momentanblindleistung Bezug L1',
    '0100180700ff': 'Momentanblindleistung Einspeisung L1',
    '01001f0600ff': 'Maximalstrom L1',
    '01001f0700ff': 'Momentanstrom L1',
    '0100200700ff': 'Momentanspannung L1',
    '0100240700ff': 'Summenwirkleistung L1',
    '0100290700ff': 'Momentanwirkleistung Bezug L2',
    '01002a0700ff': 'Momentanwirkleistung Einspeisung L2',
    '01002b0700ff': 'Momentanblindleistung Bezug L2',
    '01002c0700ff': 'Momentanblindleistung Einspeisung L2',
    '0100330600ff': 'Maximalstrom L2',
    '0100330700ff': 'Momentanstrom L2',
    '0100340700ff': 'Momentanspannung L2',
    '0100380700ff': 'Summenwirkleistung L2',
    '01003d0700ff': 'Momentanwirkleistung Bezug L3',
    '01003e0700ff': 'Momentanwirkleistung Einspeisung L3',
    '01003f0700ff': 'Momentanblindleistung Bezug L3',
    '0100400700ff': 'Momentanblindleistung Einspeisung L3',
    '0100470600ff': 'Maximalstrom L3',
    '0100470700ff': 'Momentanstrom L3',
    '0100480700ff': 'Momentanspannung L3',
    '01004c0700ff': 'Summenwirkleistung L3',
    '0100510701ff': 'Phasenabweichung Spannungen L1/L2',
    '0100510702ff': 'Phasenabweichung Spannungen L1/L3',
    '0100510704ff': 'Phasenabweichung Strom/Spannung L1',
    '010051070fff': 'Phasenabweichung Strom/Spannung L2',
    '010051071aff': 'Phasenabweichung Strom/Spannung L3',
    '010060320002': 'Aktuelle Chiptemperatur',
    '010060320003': 'Minimale Chiptemperatur',
    '010060320004': 'Maximale Chiptemperatur',
    '010060320005': 'Gemittelte Chiptemperatur',
    '010060320303': 'Spannungsminimum',
    '010060320304': 'Spannungsmaximum',
    '8181c78203ff': 'Hersteller-Identifikation',
    '8181c78205ff': 'Öffentlicher Schlüssel',
}



from binascii import a2b_hex

from smllib.sml_frame import EndOfSmlMsg, SmlFrame, SmlFrameSnippet


def check(f: SmlFrameSnippet, value, msg: str):
    if value is True or value is False or value is None or value is EndOfSmlMsg:
        assert f.value is value
    else:
        if isinstance(value, (tuple, list)):
            for i, _value in enumerate(value):
                assert f.value[i].value == _value
        else:
            assert f.value == value
    assert f.msg.hex() == msg


def test_get_int8():
    f = SmlFrame(b'\x52\xff')
    check(f.get_value(0), -1, '52ff')
    assert f.next_pos == 2

    f = SmlFrame(b'\x52\x03')
    check(f.get_value(0), 3, '5203')
    assert f.next_pos == 2

    f = SmlFrame(b'\x00\x52\x03')
    f.next_pos = 9999
    check(f.get_value(1), 3, '5203')
    assert f.next_pos == 3


def test_get_uint8():
    f = SmlFrame(b'\x62\xff')
    check(f.get_value(0), 255, '62ff')
    assert f.next_pos == 2

    f = SmlFrame(b'\x62\x03')
    check(f.get_value(0), 3, '6203')
    assert f.next_pos == 2

    f = SmlFrame(b'\x00\x62\xff')
    f.next_pos = 9999
    check(f.get_value(1), 255, '62ff')
    assert f.next_pos == 3


def test_get_int16():
    f = SmlFrame(b'\x53\xff\x00')
    check(f.get_value(0), -256, '53ff00')
    assert f.next_pos == 3

    f = SmlFrame(b'\x53\x00\x03')
    check(f.get_value(0), 3, '530003')
    assert f.next_pos == 3

    f = SmlFrame(b'\x00\x53\xff\x01')
    f.next_pos = 9999
    check(f.get_value(1), -255, '53ff01')
    assert f.next_pos == 4


def test_get_uint16():
    f = SmlFrame(b'\x63\xff\x00')
    check(f.get_value(0), 65280, '63ff00')
    assert f.next_pos == 3

    f = SmlFrame(b'\x63\x00\x03')
    check(f.get_value(0), 3, '630003')
    assert f.next_pos == 3

    f = SmlFrame(b'\x00\x63\xff\x01')
    f.next_pos = 9999
    check(f.get_value(1), 65281, '63ff01')
    assert f.next_pos == 4


def test_get_int32():
    f = SmlFrame(b'\x55\x00\x00\x0a\x8c')
    check(f.get_value(0), 2700, '5500000a8c')
    assert f.next_pos == 5

    f = SmlFrame(b'\x01\x01\x55\x00\x00\x0a\x8c')
    f.next_pos = 9999
    check(f.get_value(2), 2700, '5500000a8c')
    assert f.next_pos == 7


def test_get_uint32():
    f = SmlFrame(b'\x55\x00\x00\x0a\x8c')
    check(f.get_value(0), 2700, '5500000a8c')
    assert f.next_pos == 5


# def test_int():
#
#     f = SmlFrame(b'\x65\x0c\x6a\x50\xb5')
#     assert f.get_value(0) == 208294069
#
#     # too short stuff
#     with pytest.raises(InvalidBufferPos):
#         f = SmlFrame(b'\x56\x00\x04\xeb\x09')
#         assert f.get_value(0) is None
#     with pytest.raises(InvalidBufferPos):
#         f = SmlFrame(b'\x65\x0c\x6a')
#         assert f.get_value(0) is None
#
#     # now with indexes > 0
#     f = SmlFrame(b'\xaa\xbb\x56\x00\x04\xeb\x09\x6c')
#     assert f.get_value(2) == 82512236
#     f = SmlFrame(b'\x00\x52\xff')
#     assert f.get_value(1) == -1
#     f = SmlFrame(b'\x01\xff\x33\x62\x1e')
#     assert f.get_value(3) == 30
#     f = SmlFrame(b'\xab\xcd\x87\x44\x65\x0c\x6a\x50\xb5')
#     assert f.get_value(4) == 208294069
#
#     # now with stuff appended
#     f = SmlFrame(b'\x56\x00\x04\xeb\x09\x6c\x12\x34')
#     assert f.get_value(0) == 82512236
#     f = SmlFrame(b'\xaa\xbb\x52\xff\xcc\xdd')
#     assert f.get_value(2) == -1
#     f = SmlFrame(b'\x52\x62\x1e\x99')
#     assert f.get_value(1) == 30
#     f = SmlFrame(b'\x65\x0c\x6a\x50\xb5\x77\x88')
#     assert f.get_value(0) == 208294069
#
#     # first some values starting at index 0
#     f = SmlFrame(b'\x56\x00\x04\xeb\x09\x6c')
#     assert check(f.get_value(0), 82512236, '5600')
#

def test_none():
    f = SmlFrame(b'\x01\x01\x01')
    assert f.get_value().value is None
    assert f.next_pos == 1

    f.next_pos = 9999
    assert f.get_value(1).value is None
    assert f.next_pos == 2

    assert f.get_value().value is None
    assert f.next_pos == 3


def test_get_list():
    f = SmlFrame(b'\x71\x01')
    f.next_pos = 9999
    check(f._parse_msg(f.get_value(0)), [None], '7101')
    assert f.next_pos == 2

    f = SmlFrame(b'\x72\x01\x00')
    f.next_pos = 9999
    check(f._parse_msg(f.get_value(0)), [None, EndOfSmlMsg], '720100')
    assert f.next_pos == 3

    f = SmlFrame(b'\x72\x52\xff\x62\x1e')
    check(f._parse_msg(f.get_value(0)), [-1, 30], '7252ff621e')
    assert f.next_pos == 5

    f = SmlFrame(a2b_hex(
        '77070100010800ff'
        '6500000782'
        '01'
        '621e'
        '52ff'
        '59000000000dd359d6'
        '01'
        'ff'))
    check(f._parse_msg(f.get_value(0)),
          ['0100010800ff', 1922, None, 30, -1, 231954902, None],
          '77070100010800ff650000078201621e52ff59000000000dd359d601')
    assert f.next_pos == 28


def test_str():
    f = SmlFrame(b'\x07\x01\x00\x01\x08\x00\xFF')
    f.next_pos = 9999
    check(f.get_value(0), '0100010800ff', '070100010800ff')
    assert f.next_pos == 7


def test_long_str():
    f = SmlFrame(a2b_hex('8302010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102'))  # noqa: E501
    check(
        f.get_value(0),
        '010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102',
        '8302010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102')
    assert f.next_pos == 50

    f = SmlFrame(a2b_hex('8302010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102FF'))  # noqa: E501
    check(
        f.get_value(0),
        '010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102',
        '8302010203040101010101010101010101010101010101010101010101010101010101010101010101010101010101010102')
    assert f.next_pos == 50
    assert f.buffer[f.next_pos] == 0xFF


def test_bool():
    f = SmlFrame(b'\x42\x01')
    f.next_pos = 9999
    check(f.get_value(0), True, '4201')
    assert f.next_pos == 2

    f = SmlFrame(b'\x42\x00')
    check(f.get_value(0), False, '4200')
    assert f.next_pos == 2


def test_eom():
    f = SmlFrame(b'\x00')
    check(f.get_value(0), EndOfSmlMsg, '00')
    assert f.next_pos == 1

    f = SmlFrame(b'\x00\x00')
    check(f.get_value(1), EndOfSmlMsg, '00')
    assert f.next_pos == 2
